/*
Module : RelPaths.h
Purpose: Defines the interface for an C++ class to convert to and from relative paths, This is similiar to 
         how Visual Studio stores filenames in a project / workspace. This means that you can copy a 
         project to a different drive and recompile without problems.
Created: PJN / 07-04-2003
History: PJN / 28-12-2006 1. Updated copyright details
                          2. Addition of a CRELATIVEPATHS_EXT_CLASS preprocessor macro to allow the classes to 
                          be more easily added to an extension dll
                          3. Updated the code to clean compile on VC 2005
                          4. Updated the documentation to use the same style as the web site
         PJN / 28-06-2015 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2008 & later
                          3. Updated the code to require VC 2005 or later
                          4. Updated the code to be a header only implementation
                          5. Reworked the RelativeToAbsolute method to return the string return value as an
                          output parameter instead of a return value.
                          6. Updated the code to include SAL annotations.
                          7. Reworked the code to optionally compile without MFC. By default the code now use 
                          STL classes and idioms but if you define RELATIVEPATHS_MFC_EXTENSIONS the code will 
                          revert back to the MFC behaviour.
                          8. Removed the static critical section member variable of the CRelativePaths class. 
                          Please note that while the RelativeToAbsolute  method is executing it is possible 
                          that code in some other thread of your app could also be modifying the current 
                          working directory. To avoid such race conditions you should serialize that code 
                          with the execution of RelativeToAbsolute with something like a critical section.
         PJN / 15-12-2018 1. Updated copyright details
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
                          3. Removed code path which supported RELATIVEPATHS_MFC_EXTENSIONS define
         PJN / 17-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 15-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 2003 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


////////////////////// Macros / Defines ///////////////////////////////////////

#pragma once

#ifndef __RELPATHS_H__
#define __RELPATHS_H__

#ifndef CRELATIVEPATHS_EXT_CLASS
#define CRELATIVEPATHS_EXT_CLASS
#endif

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef __ATLDEF_H__
#pragma message("To avoid this message, please put atldef.h in your pre compiled header (normally stdafx.h)")
#include <atldef.h>
#endif //#ifndef __ATLDEF_H__

#ifndef __ATLSTR_H__
#pragma message("To avoid this message, please put atlstr.h in your pre compiled header (normally stdafx.h)")
#include <atlstr.h>
#endif //#ifndef __ATLSTR_H__


////////////////////// Classes ////////////////////////////////////////////////

class CRELATIVEPATHS_EXT_CLASS CRelativePaths
{
public:
//Typedefs
#ifdef _UNICODE
  using String = std::wstring;
#else
  using String = std::string;
#endif //#ifdef _UNICODE

//Methods
  _Success_(return != false) static bool AbsoluteToRelative(_In_ const String& sAbsolutePath, _In_ const String& sRelativeTo, _Out_ String& sRelativePath)
  {
    //Set the output parameter to a reasonable default
    sRelativePath.clear();

    //Break up the absolute path into its constituent parts
    ATL::CAtlString sAbsoluteDrive;
    ATL::CAtlString sAbsoluteDir;
    ATL::CAtlString sAbsoluteFname;
    ATL::CAtlString sAbsoluteExt;
    _tsplitpath_s(sAbsolutePath.c_str(), sAbsoluteDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sAbsoluteDir.GetBuffer(_MAX_DIR), _MAX_DIR,
                  sAbsoluteFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, sAbsoluteExt.GetBuffer(_MAX_EXT), _MAX_EXT);
    sAbsoluteExt.ReleaseBuffer();
    sAbsoluteFname.ReleaseBuffer();
    sAbsoluteDir.ReleaseBuffer();
    sAbsoluteDrive.ReleaseBuffer();

    //Break up the "sRelativeTo" path into its constituent parts
    ATL::CAtlString sRelativeToDrive;
    ATL::CAtlString sRelativeToDir;
    _tsplitpath_s(sRelativeTo.c_str(), sRelativeToDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sRelativeToDir.GetBuffer(_MAX_DIR), _MAX_DIR, nullptr, 0, nullptr, 0);
    sRelativeToDir.ReleaseBuffer();
    sRelativeToDrive.ReleaseBuffer();

    //We can only convert an absolute path to a relative path if the paths are on the same drive or UNC machine
    if (sAbsoluteDrive.CompareNoCase(sRelativeToDrive) == 0)
    {
      bool bMatch{false};
      int nRelativeUpLevel{0};
      ATL::CAtlString sRelativeAncestor;
      while (!bMatch)
      {
        sRelativeAncestor = GetFolderAncestor(sRelativeToDir, nRelativeUpLevel);
        if (sRelativeAncestor != _T("\\"))
        {
#pragma warning(suppress: 26477)
          ATLASSERT(sRelativeAncestor.GetLength());
          if (sAbsoluteDir.Find(sRelativeAncestor) == 0)
          {
            bMatch = true;
            nRelativeUpLevel = GetAncestorLevel(sRelativeToDir, sRelativeAncestor);

            for (int i{0}; i<nRelativeUpLevel; i++)
#pragma warning(suppress: 26489)
              sRelativePath += _T("..\\");

            const int nSubLength{sAbsoluteDir.GetLength() - sRelativeAncestor.GetLength() - 1};
#pragma warning(suppress: 26489)
            sRelativePath += sAbsoluteDir.Mid(sAbsoluteDir.GetLength() - nSubLength, nSubLength);
#pragma warning(suppress: 26489)
            sRelativePath += sAbsoluteFname;
#pragma warning(suppress: 26489)
            sRelativePath += sAbsoluteExt;
          }
          else
            nRelativeUpLevel++;
        }
        else
        {
          //No matches between the absolute and relative path in that case then the number of levels up
          //for the relative path corresponds to the number of directory levels in sRelativeTo
          bMatch = true;
          nRelativeUpLevel = FoldersUpToRootLevel(sRelativeTo.c_str());

          for (int i{0}; i<nRelativeUpLevel; i++)
#pragma warning(suppress: 26489)
            sRelativePath += _T("..\\");

          const int nSubLength{sAbsoluteDir.GetLength() - sRelativeAncestor.GetLength()};
#pragma warning(suppress: 26489)
          sRelativePath += sAbsoluteDir.Mid(sAbsoluteDir.GetLength() - nSubLength, nSubLength);
#pragma warning(suppress: 26489)
          sRelativePath += sAbsoluteFname;
#pragma warning(suppress: 26489)
          sRelativePath += sAbsoluteExt;
        }
      }
    }
    else
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return false;
    }

    return true;
  }

  _Success_(return != false) static bool RelativeToAbsolute(_In_ const String& sRelativePath, _In_ const String& sRelativeTo, _Out_ String& sAbsolutePath)
  {
    //Break up the relativeto path into its constituent parts
    ATL::CAtlString sRelativeToDrive;
    ATL::CAtlString sRelativeToDir;
    _tsplitpath_s(sRelativeTo.c_str(), sRelativeToDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sRelativeToDir.GetBuffer(_MAX_DIR), _MAX_DIR, nullptr, 0, nullptr, 0);
    sRelativeToDir.ReleaseBuffer();
    sRelativeToDrive.ReleaseBuffer();
    ATL::CAtlString sRelativeToDirectory;
    _tmakepath_s(sRelativeToDirectory.GetBuffer(_MAX_PATH), _MAX_PATH, sRelativeToDrive, sRelativeToDir, nullptr, nullptr);
    sRelativeToDirectory.ReleaseBuffer();

    //Remove any trailing "\"
    const auto nRealtiveToDirectoryLen{sRelativeToDirectory.GetLength() - 1};
    if ((nRealtiveToDirectoryLen > 2) && sRelativeToDirectory[nRealtiveToDirectoryLen] == _T('\\'))
      sRelativeToDirectory = sRelativeToDirectory.Left(nRealtiveToDirectoryLen);

    //Only makes sense to convert relative to absolute if sRelativeTo contains drive letter or is a UNC path
    if (sRelativeToDrive.GetLength() || (sRelativeToDir.Find(_T("\\\\")) == 0))
    {
      //Remember the old current directory before we change it
      ATL::CAtlString sOldCurrentDirectory;
      const DWORD dwGCD{GetCurrentDirectory(_MAX_PATH, sOldCurrentDirectory.GetBuffer(_MAX_PATH))};
      sOldCurrentDirectory.ReleaseBuffer();
      if (dwGCD == 0)
        return false;

      //Temporarily set the current directory to the relativeTo directory
      if (SetCurrentDirectory(sRelativeToDirectory))
      {
        ATL::CAtlString sTempAbsolutePath;

        //Convert the relative path to an absolute path
        LPTSTR szFilePart{nullptr};
        const DWORD dwGFPN{GetFullPathName(sRelativePath.c_str(), _MAX_PATH, sTempAbsolutePath.GetBuffer(_MAX_PATH), &szFilePart)};
        sTempAbsolutePath.ReleaseBuffer();
        if (dwGFPN == 0)
          return false;
        sAbsolutePath = sTempAbsolutePath;

        //Restore the current working directory
        SetCurrentDirectory(sOldCurrentDirectory);
      }
      else
        return false;
    }
    else
      sAbsolutePath = sRelativePath;

    return true;
  }

  static ATL::CAtlString GetFolderAncestor(_In_z_ LPCTSTR pszDir, _In_ int nLevel)
  {
    //What will be the return value
    ATL::CAtlString sAncestor{pszDir};

    //Remove the trailing "\" if any
    const auto nAncestorLength{sAncestor.GetLength()};
#pragma warning(suppress: 26477)
    ATLASSERT(nAncestorLength);
#pragma warning(suppress: 26446)
    if ((nAncestorLength > 1) && sAncestor[nAncestorLength - 1] == _T('\\'))
      sAncestor = sAncestor.Left(nAncestorLength - 1); 

    //Find the specified amount of folders up
    for (int i{0}; i<nLevel && sAncestor.GetLength(); i++)
    {
      const auto nIndex{sAncestor.ReverseFind(_T('\\'))};
      if (nIndex != -1)
      {
        sAncestor = sAncestor.Mid(0, nIndex);
        if ((_tcscmp(pszDir, _T("\\")) != 0) && (i == (nLevel - 1)) && (sAncestor.GetLength() == 0))
          sAncestor = _T("\\");
      }
    }

    return sAncestor;
  }

  static int FoldersUpToRootLevel(_In_z_ LPCTSTR pszDir)
  {
    ATL::CAtlString sAncestor{pszDir};

    //Remove the trailing "\" if any
    const auto nAncestorLength{sAncestor.GetLength()};
#pragma warning(suppress: 26477)
    ATLASSERT(nAncestorLength);
#pragma warning(suppress: 26446)
    if (sAncestor[nAncestorLength - 1] == _T('\\'))
      sAncestor = sAncestor.Mid(0, nAncestorLength - 1);

    int nLevels{-1};
    while (sAncestor.GetLength())
    {
      const auto nIndex{sAncestor.ReverseFind(_T('\\'))};
      if (nIndex != -1)
      {
        sAncestor = sAncestor.Mid(0, nIndex);
        nLevels++;
      }
      else
        sAncestor.Empty();
    }

    return nLevels;
  }

  static int GetAncestorLevel(_In_z_ LPCTSTR pszDir, _In_z_ LPCTSTR pszSubDir)
  {
    //By default assume failure
    int nLevel{-1};

    //What will be the return value
    ATL::CAtlString sLocalDir{pszDir};

    if (sLocalDir.Find(pszSubDir) == 0)
      nLevel = FoldersUpToRootLevel(pszDir) - FoldersUpToRootLevel(pszSubDir);

    return nLevel;
  }
};

#endif //#ifndef __RELPATHS_H__
